from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
import datetime
import functools
import hashlib
import json
import logging
import os
import pickle
import shutil


# TODO: use stricter structure
_EntryDict = Dict
_IndexDict = Dict[str, _EntryDict]


_CACHE_INDEX_FILE = 'index.json'
_DEFAULT_CACHE_DIR = '.derpcache/'
_CACHE_CONFIG_DEFAULTS = {
    'cache_dir': _DEFAULT_CACHE_DIR,
}


logger = logging.getLogger(__name__)


__cache_config = _CACHE_CONFIG_DEFAULTS.copy()


def update_cache_config(**config) -> dict:
    """Update cache config settings.

    Args:

        config (:obj:`dict`): Dictionary of configuration settings.

            Currently supported keys:

                "cache_dir": (path to) the desired cache directory.

    Returns:

        dict: The current configuration settings.
    """

    __cache_config.update(config)
    return __cache_config


def reset_cache_config() -> dict:
    """Resets config settings to package defaults.

    Returns:

        dict: The default configuration settings.
    """

    for k in list(__cache_config.keys()):
        __cache_config.pop(k)
    return update_cache_config(**_CACHE_CONFIG_DEFAULTS)


def _get_cache_dir() -> str:
    return __cache_config['cache_dir']


def _get_cache_path(filename: str = '') -> str:
    return os.path.join(_get_cache_dir(), filename)


def _get_index_path() -> str:
    return _get_cache_path(_CACHE_INDEX_FILE)


def _is_non_str_iterable(x: Any) -> bool:
    return hasattr(x, '__iter__') and not isinstance(x, str)


def _sort_nested_dicts(value: Union[dict, list, Any]) -> Union[dict, list, Any]:
    """Sort nested dicts by keys so casting it will produce a deterministic string.

    Warning: Thar be edge cases."""

    if isinstance(value, dict):
        value = {k: _sort_nested_dicts(v) for k, v in sorted(value.items(), key=str)}
    elif _is_non_str_iterable(value):
        value = tuple(_sort_nested_dicts(x) for x in value)
    return value


def _to_string(arg: Any) -> str:
    arg = _sort_nested_dicts(arg)
    return str(arg)


def _hash_args(*args, **kwargs) -> str:
    args_string = _to_string(args) + _to_string(kwargs)
    return hashlib.sha256(args_string.encode()).hexdigest()[:8]


def _read_index() -> _IndexDict:
    with open(_get_index_path(), 'r') as f:
        index = json.load(f)
    return index


def _write_index(index: _IndexDict) -> None:
    with open(_get_index_path(), 'w') as f:
        json.dump(index, f)


def _write_entry_to_index(
    index: _IndexDict,
    hash: str,
    entry: _EntryDict,
) -> _IndexDict:
    index[hash] = entry
    _write_index(index)
    return index


def get_by_hash(hash: str) -> Any:
    """Retrieve the function call's return value by its hash.

    Args:

        hash (:obj:`str`): The hash of the function call.

    Returns:

        Any: The return value of the function call.
    """

    with open(_get_cache_path(hash), 'rb') as f:
        value = pickle.load(f)
    return value


def _write_object_by_hash(hash: str, value: Any) -> None:
    with open(_get_cache_path(hash), 'wb') as f:
        pickle.dump(value, f)


def _remove_objects(to_remove: List[str]) -> None:
    for hash in to_remove:
        os.remove(_get_cache_path(hash))


def _remove_entries(index: _IndexDict, to_remove: List[str]) -> _IndexDict:
    index = {k: v for k, v in index.items() if k not in to_remove}
    _write_index(index)
    return index


def _is_expired(entry: _EntryDict) -> bool:
    expires_after = entry.get('expires_after')
    if expires_after:
        called_at = datetime.datetime.fromisoformat(entry['called_at'])
        expires_after = datetime.timedelta(seconds=expires_after)
        now = datetime.datetime.utcnow()
        expired = called_at + expires_after < now
    else:
        expired = False
    return expired


def _remove_expired_items(index: _IndexDict) -> _IndexDict:
    to_remove = []
    for hash, entry in index.items():
        if _is_expired(entry):
            to_remove.append(hash)
    index = _remove_entries(index, to_remove)
    _remove_objects(to_remove)
    return index


def _sort_index(index: _IndexDict) -> _IndexDict:
    index = {k: v for k, v in sorted(index.items(), key=lambda x: x[1]['called_at'])}
    return index


def get_index(clear_expired: bool = True) -> _IndexDict:
    """Retrieve `index.json` metadata dict about cache contents.

    Note: When using `expires_after` expiration rules, expired cache contents will be
        permanently removed upon calling this function (or checking the cache again),
        unless `clear_expired` is set to `False`.

    Args:
        clear_expired (bool): Clear expired cache contents upon call.

    Returns:
        dict: The current state of the cache.
    """

    index = _read_index()
    if clear_expired:
        index = _remove_expired_items(index)
    index = _sort_index(index)
    return index


def _init_cache() -> None:
    try:
        os.makedirs(_get_cache_dir())
        _write_index({})
    except FileExistsError:
        pass


def clear_cache() -> None:
    """Removes cache directory and all files within it.  If configured cache directory
    is a path, remove only the bottom-most empty directories within that path.
    """

    def _remove_bottom_dir(path):
        dirs = path.rstrip('/').split('/')
        new_path = '/'.join(dirs[:-1])
        return new_path

    cache_path = _get_cache_path()
    shutil.rmtree(cache_path, ignore_errors=True)
    cache_path = _remove_bottom_dir(cache_path)
    while cache_path:
        try:
            os.rmdir(cache_path)
            cache_path = _remove_bottom_dir(cache_path)
        except OSError:
            break


def _describe_callable(f: Callable) -> str:
    """Note: Some callables are missing a :attr:`__qualname__`, so including `type()`
    provides at least some information."""

    mod = f.__module__
    name = getattr(f, '__qualname__', str(type(f)))
    return f'{mod}.{name}'


def _expires_after_to_float(expires_after: Union[float, datetime.timedelta]) -> float:
    if isinstance(expires_after, datetime.timedelta):
        expires_after = expires_after.total_seconds()
    return expires_after


def _format_entry(
    f: Callable,
    called_at: str,
    expires_after: Optional[Union[float, datetime.timedelta]],
    annotation: Optional[str],
) -> _EntryDict:
    entry = {
        'callable': _describe_callable(f),
        'called_at': called_at,
    }
    if expires_after:
        # mypy thinks `expires_after` is a string here
        entry['expires_after'] = _expires_after_to_float(expires_after)  # type: ignore
    if annotation:
        entry['annotation'] = annotation
    return entry


def cache(
    f: Callable,
    *args,
    _expires_after: Optional[Union[float, datetime.timedelta]] = None,
    _annotation: Optional[str] = None,
    **kwargs,
) -> Any:
    """
    Calls a function and caches the results.

    Args:

        f (Callable):

            Function whose results are to be cached.

        *args (optional):

            The function call's arguments.

        **kwargs (optional):

            The function call's keyword arguments.

        _expires_after (float, :obj:`datetime.timedelta`, optional):

            Time elapsed after which the cache entry will be cleared.  Numeric values
                are interpreted as seconds.

        _annotation (:obj:`str`, optional):

            Arbitrary string that can be passed to help identify or describe the call.

    Returns:

        value (any):

            The return value of the original function call.
    """

    _init_cache()
    hash = _hash_args(
        _describe_callable(f),  # lazy, but keeps :meth:`_hash_args` dumb
        *args,
        **kwargs,
    )
    index = get_index(clear_expired=True)
    if hash in index:
        value = get_by_hash(hash)
        logger.debug('cache hit')
    else:
        logger.debug('caching...')
        called_at = datetime.datetime.utcnow().isoformat()
        value = f(*args, **kwargs)
        _write_object_by_hash(hash, value)
        _write_entry_to_index(
            index,
            hash,
            entry=_format_entry(
                f,
                called_at,
                _expires_after,
                _annotation,
            ),
        )
        logger.debug('caching successful.')
    return value


def cache_wrapper(
    _expires_after: Optional[Union[float, datetime.timedelta]] = None,
    _annotation: Optional[str] = None,
) -> Callable:

    # TODO: support wrapping bound methods.

    def decorator(f: Callable) -> Callable:
        @functools.wraps(f)
        def wrapped(*args, **kwargs) -> Any:
            return cache(
                f,
                *args,
                **kwargs,
                _expires_after=_expires_after,
                _annotation=_annotation,
            )

        return wrapped

    return decorator



from derpcache import _cache
from faker import Faker
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Union
import datetime
import logging
import os
import pytest


faker = Faker()


_RandomValueUnion = Union[str, int, float]
_RandomDepthDict = Dict


@pytest.fixture(autouse=True)
def _auto_clear_cache():
    yield
    _cache.clear_cache()
    _cache.reset_cache_config()


@pytest.fixture(autouse=True)
def _auto_caplog(caplog):
    with caplog.at_level(logging.INFO):
        yield


def _randomize_value() -> _RandomValueUnion:
    return faker.random_element((faker.lexify, faker.pyint, faker.pyfloat))()


def _randomize_args() -> Tuple[_RandomValueUnion, ...]:
    return tuple(_randomize_value() for _ in range(faker.pyint(2, 4)))


def _randomize_kwargs(depth: int = 1) -> _RandomDepthDict:
    max_nested_dict_depth = 3
    if depth == 1:
        k = faker.lexify()
    else:
        k = _randomize_value()
    if faker.pybool() and not depth > max_nested_dict_depth:
        v = _randomize_kwargs(depth=depth + 1)
    else:
        # mypy has trouble with random recursive dict creation
        v = _randomize_value()  # type: ignore
    return {k: v for _ in range(faker.pyint(1, 3))}


def test__sort_nested_dicts():
    from collections import OrderedDict  # order matters for comparison

    d = OrderedDict(
        {
            'b': OrderedDict(
                {
                    'e': 1,
                    'f': [1, 2, [], {}],
                    'd': [
                        OrderedDict({3: 'i', 1: 'g', 2: 'h'}),
                        1,
                        OrderedDict({'k': 2, 'j': [1, 2]}),
                    ],
                }
            ),
            'a': OrderedDict(
                {
                    'm': [1, OrderedDict({'p': [1, 2], 'o': {}})],
                    'l': 1,
                    'n': OrderedDict({'r': 2, 'q': 1}),
                }
            ),
        }
    )
    expected_d = OrderedDict(
        {
            'a': OrderedDict(
                {
                    'l': 1,
                    'm': (1, OrderedDict({'o': {}, 'p': (1, 2)})),
                    'n': OrderedDict({'q': 1, 'r': 2}),
                }
            ),
            'b': OrderedDict(
                {
                    'd': (
                        OrderedDict({1: 'g', 2: 'h', 3: 'i'}),
                        1,
                        OrderedDict({'j': (1, 2), 'k': 2}),
                    ),
                    'e': 1,
                    'f': (1, 2, (), {}),
                }
            ),
        }
    )
    actual_d = _cache._sort_nested_dicts(d)
    assert actual_d == expected_d


class Test__make_hash:
    @pytest.mark.parametrize('args', [(), _randomize_args()])
    @pytest.mark.parametrize('kwargs', [{}, _randomize_kwargs()])
    def test__make_hash__different_orders(self, args, kwargs):
        def _reverse_nested_dicts(
            value: Union[dict, list, Any]
        ) -> Union[dict, list, Any]:
            if isinstance(value, dict):
                value = {
                    k: _reverse_nested_dicts(v) for k, v in reversed(value.items())
                }
            elif _cache._is_non_str_iterable(value):
                value = [_reverse_nested_dicts(x) for x in value]
            return value

        args1, kwargs1 = args, kwargs
        args2 = args
        kwargs2 = _reverse_nested_dicts(kwargs)

        hash1 = _cache._hash_args(*args1, **kwargs1)
        hash2 = _cache._hash_args(*args2, **kwargs2)

        assert hash1 == hash2

    @pytest.mark.parametrize(
        'differing_args',
        [
            ('args'),
            ('kwargs'),
            ('args', 'kwargs'),
        ],
    )
    def test__make_hash__different_args(self, differing_args):
        args1 = _randomize_args()
        args2 = _randomize_args() if 'args' in differing_args else (args1)
        kwargs1 = _randomize_kwargs()
        kwargs2 = _randomize_kwargs() if 'kwargs' in differing_args else kwargs1
        hash1 = _cache._hash_args(*args1, **kwargs1)
        hash2 = _cache._hash_args(*args2, **kwargs2)
        assert hash1 != hash2


def _func1(*args, **kwargs) -> _RandomValueUnion:
    logging.info(f'test func1 called with args: {args} and kwargs: {kwargs}')
    return _randomize_value()


def _func2(*args, **kwargs) -> _RandomValueUnion:
    logging.info(f'test func2 called with args: {args} and kwargs: {kwargs}')
    return _randomize_value()


class Test__cache:
    def test__cache__get_index(self, caplog, freezer):
        dt = faker.date_time()
        freezer.move_to(dt)
        _cache._init_cache()
        index1 = _cache.get_index()

        assert index1 == {}

        _cache.cache(_func1)
        index2 = _cache.get_index()

        assert len(index2) == 1
        ((_, entry),) = index2.items()
        expected_entry = {
            'callable': _cache._describe_callable(_func1),
            'called_at': dt.isoformat(),
        }
        assert entry == expected_entry

    def test__cache__hit(self, caplog):
        with caplog.at_level(logging.INFO):
            result1 = _func1()
            result2 = _func1()

        assert len(caplog.messages) == 2
        assert result1 != result2

        caplog.clear()

        result3 = _cache.cache(_func1)
        result4 = _cache.cache(_func1)

        assert len(caplog.messages) == 1
        assert result3 == result4

    def test__cache__with_args(self, caplog):
        args1, kwargs1 = _randomize_args(), _randomize_kwargs()
        args2, kwargs2 = _randomize_args(), _randomize_kwargs()

        args1_result1 = _cache.cache(_func1, *args1, **kwargs1)
        args1_result2 = _cache.cache(_func1, *args1, **kwargs1)
        args2_result1 = _cache.cache(_func1, *args2, **kwargs2)

        assert args1_result1 == args1_result2
        assert args1_result1 != args2_result1
        assert len(caplog.messages) == 2

    def test__cache__with_annotation__unhashed_annotation(self, caplog):
        args1, kwargs1 = _randomize_args(), _randomize_kwargs()
        args2, kwargs2 = _randomize_args(), _randomize_kwargs()
        annotation = faker.lexify()

        result1 = _cache.cache(_func1, *args1, **kwargs1, _annotation=annotation)
        result2 = _cache.cache(_func1, *args1, **kwargs1)
        result3 = _cache.cache(_func1, *args2, **kwargs2)

        assert result1 == result2
        assert result1 != result3
        index = _cache.get_index()
        assert len(index) == 2
        (index_entry, _) = index.values()
        assert index_entry['annotation'] == annotation
        assert len(caplog.messages) == 2

    def test__cache__clear_cache(self, caplog):
        result1 = _cache.cache(_func1)
        result2 = _cache.cache(_func1)

        assert result1 == result2
        assert len(caplog.messages) == 1

        _cache.clear_cache()
        caplog.clear()

        result3 = _cache.cache(_func1)
        result4 = _cache.cache(_func1)

        assert result3 == result4
        assert result1 != result3
        assert len(caplog.messages) == 1

    def test__cache__multiple_entries(self, caplog):
        func1_result1 = _cache.cache(_func1)
        func1_result2 = _cache.cache(_func1)
        func2_result1 = _cache.cache(_func2)
        func2_result2 = _cache.cache(_func2)

        assert func1_result1 == func1_result2
        assert func2_result1 == func2_result2
        assert func1_result1 != func2_result1
        index = _cache.get_index()
        assert len(index) == 2
        assert len(caplog.messages) == 2

    def test__get_by_hash(self, caplog, freezer):
        dt = faker.date_time()
        freezer.move_to(dt)
        annotation = faker.lexify()

        result1 = _cache.cache(
            _func1,
            *_randomize_args(),
            **_randomize_kwargs(),
            _annotation=annotation,
        )

        index1 = _cache.get_index()
        assert len(index1) == 1
        ((hash1, entry1),) = index1.items()
        expected_entry1 = {
            'callable': _cache._describe_callable(_func1),
            'called_at': dt.isoformat(),
            'annotation': annotation,
        }
        assert entry1 == expected_entry1
        assert _cache.get_by_hash(hash1) == result1
        assert len(caplog.messages) == 1

        later = dt + datetime.timedelta.resolution
        freezer.move_to(later)
        result2 = _cache.cache(_func2, *_randomize_args(), **_randomize_kwargs())

        assert result2 != result1
        index2 = _cache.get_index()
        assert len(index2) == 2
        (_, _), (hash2, entry2) = index2.items()
        assert hash2 != hash1
        expected_entry2 = {
            'callable': _cache._describe_callable(_func2),
            'called_at': later.isoformat(),
        }
        assert entry2 == expected_entry2
        assert _cache.get_by_hash(hash2) == result2
        assert len(caplog.messages) == 2

    @pytest.mark.parametrize(
        'cache_dir',
        [
            faker.lexify('????/'),
            faker.lexify('????/????/'),
        ],
    )
    def test__cache__set_cache_dir(self, caplog, cache_dir):
        config = _cache.update_cache_config(cache_dir=cache_dir)
        assert config == {'cache_dir': cache_dir}

        with caplog.at_level(logging.INFO):
            result1 = _cache.cache(_func1)
            result2 = _cache.cache(_func1)

        assert len(caplog.messages) == 1
        assert result1 == result2

        assert _cache._DEFAULT_CACHE_DIR.rstrip('/') not in os.listdir('.')
        cache_contents = os.listdir(cache_dir)
        assert len(cache_contents) == 2
        index = _cache.get_index()
        assert len(index) == 1

        _cache.clear_cache()
        caplog.clear()
        assert cache_dir.split('/')[0] not in os.listdir('.')

        with caplog.at_level(logging.INFO):
            result3 = _cache.cache(_func1)
            result4 = _cache.cache(_func1)

        assert len(caplog.messages) == 1
        assert result3 == result4
        assert result3 != result1

        assert _cache._DEFAULT_CACHE_DIR.rstrip('/') not in os.listdir('.')
        cache_contents = os.listdir(cache_dir)
        assert len(cache_contents) == 2
        index = _cache.get_index()
        assert len(index) == 1

        _cache.clear_cache()
        caplog.clear()
        assert cache_dir.split('/')[0] not in os.listdir('.')

        config = _cache.reset_cache_config()

        with caplog.at_level(logging.INFO):
            result5 = _cache.cache(_func1)
            result6 = _cache.cache(_func1)

        assert len(caplog.messages) == 1
        assert result5 == result6
        assert result5 != result3

        assert _cache._DEFAULT_CACHE_DIR.rstrip('/') in os.listdir('.')


@pytest.mark.parametrize('expires_after_type', (float, datetime.timedelta))
def test__expires_after__clear_expired(caplog, freezer, expires_after_type):
    dt_called = faker.date_time()
    freezer.move_to(dt_called)
    args, kwargs = _randomize_args(), _randomize_kwargs()
    expires_after_float = faker.pyfloat(min_value=1, max_value=3600, right_digits=6)
    expires_after_delta = datetime.timedelta(seconds=expires_after_float)
    if expires_after_type == float:
        expires_after = expires_after_float
    else:
        expires_after = expires_after_delta

    result = _cache.cache(_func1, *args, **kwargs, _expires_after=None)
    result_expires = _cache.cache(_func2, *args, **kwargs, _expires_after=expires_after)

    index1 = _cache.get_index(clear_expired=False)
    assert len(index1) == 2
    assert len(caplog.messages) == 2
    (
        (hash, entry),
        (hash_expires, entry_expires),
    ) = index1.items()
    expected_entry = {
        'callable': _cache._describe_callable(_func1),
        'called_at': dt_called.isoformat(),
    }
    assert entry == expected_entry
    assert result == _cache.get_by_hash(hash)
    expected_entry_expires = {
        'callable': _cache._describe_callable(_func2),
        'called_at': dt_called.isoformat(),
        'expires_after': expires_after_float,
    }
    assert entry_expires == expected_entry_expires
    assert result_expires == _cache.get_by_hash(hash_expires)

    dt_expired = dt_called + expires_after_delta + datetime.timedelta.resolution
    freezer.move_to(dt_expired)
    index2 = _cache.get_index(clear_expired=True)

    assert len(index2) == 1
    assert len(caplog.messages) == 2
    assert index2 == {hash: entry}
    assert hash_expires not in index2


def test__cache_wrapper():
    pass



from ._cache import cache
from ._cache import cache_wrapper
from ._cache import clear_cache
from ._cache import get_by_hash
from ._cache import get_index


"""
derpcache

'A simple pickle-based caching utility.'
"""


__version__ = '0.0.1'
__author__ = 'Ben Johnson'
__credits__ = 'Silver Zinc Beetle'
__all__ = [
    'cache',
    'cache_wrapper',
    'clear_cache',
    'get_index',
    'get_by_hash',
]



from setuptools import find_packages
from setuptools import setup


with open('README.md', 'r') as f:
    long_description = f.read()


setup(
    name='derpcache',
    version='0.1.0',
    description='A simple pickle-based caching utility.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='GPL-3',
    author='Ben Johnson',
    author_email='bkitej@gmail.com',
    packages=find_packages(),
    install_requires=[],  # Nothing!
    extras_require={
        'dev': [
            'commitizen',
            'pre-commit',
        ],
        'lint': [
            'black',
            'flakeheaven',
            'mypy',
            'types-setuptools',
        ],
        'test': [
            'faker',
            'pytest',
            'pytest-freezegun',
        ],
    },
    zip_safe=False,  # for mypy to detect typing stubs
)
